<?php
require_once(dirname(__FILE__).'/../../../config.php');
require_once(dirname(__FILE__).'/../../lib.php');

// idnumber of magic course
define('SHAREDACTV_MAGIC_COURSE','!sharedactivities');

// shortnames of participant roles
define('SHAREDACTV_ROLE_PARTICIPANT','sharedactv_participant');
define('SHAREDACTV_ROLE_CREATOR','sharedactv_creator');

// enrol type for roles assigned by this system
define('SHAREDACTV_ENROLTYPE','sharedactv');

// OU shared APIs which (for OU system) are present in local, elsewhere
// are incorporated in module
@include_once(dirname(__FILE__).'/../../../local/transaction_wrapper.php');
if (!class_exists('transaction_wrapper')) {
    require_once(dirname(__FILE__).'/null_transaction_wrapper.php');
}

/**
 * Obtains the 'magic course' used to store shared activities. If it does not
 * already exist, the magic course is created and initialised. Always returns
 * the course object if it returns; calls 'error' in event of failure.
 * @return The course object
 */
function sharedactv_get_magic_course() {
    // Obtain course
    $course=get_record('course','idnumber',SHAREDACTV_MAGIC_COURSE);

    // If it doesn't exist, create it
    if(!$course) {
        global $CFG;
        require_once($CFG->dirroot.'/course/lib.php');

        // Put it in first category
        if(!($categories=get_records('course_categories','','','id','id',0,1))) {
            error('Error obtaining arbitrary course category');
        }

        // Set up course information
        $data=new StdClass;
        $data->shortname=get_string('formatsharedactv','format_sharedactv');
        $data->idnumber=SHAREDACTV_MAGIC_COURSE;
        $data->fullname=get_string('fullname','format_sharedactv');
        $data->category=reset($categories)->id;
        $data->format='sharedactv';
        $data->visible=1;
        if(!($course=create_course($data))) {
            error('Error creating magic course');
        }

        // Set it up so all users can access
        if(!($context=get_context_instance(CONTEXT_COURSE,$course->id))) {
            error('Error obtaining context for magic course');
        }
        if(empty($CFG->defaultuserroleid)) {
            error('Default user role not defined, check settings under Users/Permissions/User Policies');
        }
        if(!assign_capability('moodle/course:view',1,$CFG->defaultuserroleid,$context->id)) {
        		error('Failed to set up magic course role overrides');
        }
    }

    return $course;
}

/**
 * @param object $course Course object, must have ->idnumber
 * @return bool True if this is the shared-activity course 
 */
function sharedactv_is_magic_course($course) {
    return $course->idnumber===SHAREDACTV_MAGIC_COURSE;
}

/**
 * Obtains the role IDs of the two roles for participants in a shared activity.
 * If these roles do not exist, they are created. Caches results within request.
 * @return array 2-element array with SHAREDACTV_ROLE_xx constant as key and
 *   role id (integer) as value
 */
function sharedactv_get_participant_roles() {
    static $results=array();
    if(count($results)==2) {
        return $results;
    }

    // Obtain both IDs using one query
    global $CFG;
    $rs=get_recordset_sql("SELECT shortname,id FROM {$CFG->prefix}role WHERE shortname IN ('".
        SHAREDACTV_ROLE_PARTICIPANT."','".SHAREDACTV_ROLE_CREATOR."')");
    if(!$rs) {
        error('Failed to contact database to obtain shared activity roles');
    }
    while($rec=rs_fetch_next_record($rs)) {
        $results[$rec->shortname]=$rec->id;
    }
    rs_close($rs);

    // Did we not get them? Maybe they need creating.
    if(!array_key_exists(SHAREDACTV_ROLE_PARTICIPANT,$results)) {
        $results[SHAREDACTV_ROLE_PARTICIPANT]=
            sharedactv_create_role(SHAREDACTV_ROLE_PARTICIPANT);
    }
    if(!array_key_exists(SHAREDACTV_ROLE_CREATOR,$results)) {
        $results[SHAREDACTV_ROLE_CREATOR]=
            sharedactv_create_role(SHAREDACTV_ROLE_CREATOR);
    }
    if(count($results)>2) {
        error('Unexpected roles in database that match the shared activities shortnames');
    }

    // Definitely got them now
    return $results;
}

/**
 * Creates a role for the shared activity system. Initially the role
 * permissions are set up the same as for legacy student. For the special
 * case of SHAREDACTV_ROLE_CREATOR, permissions are set up as for legacy
 * non-editing teacher. Both can be configured after creation.
 * @param string $shortname
 * @return int ID of new role
 */
function sharedactv_create_role($shortname) {
    global $CFG;
    $sitecontext = get_context_instance(CONTEXT_SYSTEM);

    $tw=new transaction_wrapper();

    // Create role entry
    $newrole=new StdClass;
    $newrole->shortname=$shortname;
    $newrole->name=get_string($shortname,'format_sharedactv');
    $newrole->description=get_string($shortname.'_desc','format_sharedactv');
    $newrole->sortorder=get_field_sql("SELECT MAX(sortorder)+1 FROM {$CFG->prefix}role");
    if(!($id=insert_record('role',$newrole))) {
        $tw->rollback();
        error('Failed to create role for shared activity system');
    }

    // Set up as default student, or non-editing teacher if using the creator
    // role
    $legacyroles=get_legacy_roles();
    if(!(assign_capability(
        $legacyroles[$shortname===SHAREDACTV_ROLE_CREATOR ? 'teacher' : 'student'],
        CAP_ALLOW,$id,$sitecontext->id))) {
        $tw->rollback();
        error('Failed to assign legacy role capability for shared activity role');
    }

    // Reset capabilities to include all the default ones for that type of
    // legacy role
    reset_role_capabilities($id); // Does not return a value so cannot check error

    // Now include only those that apply at module level
    if(!execute_sql("DELETE FROM {$CFG->prefix}role_capabilities
        WHERE roleid=$id AND NOT (capability LIKE 'mod/%')",false)) {
        $tw->rollback();
        error('Failed to delete unnecessary capabilities for shared activity role');
    }
    
    // Add special creator role to allow managing activities
    if($shortname===SHAREDACTV_ROLE_CREATOR) {
        if(!assign_capability('format/sharedactv:manage',CAP_ALLOW,$id,$sitecontext->id)) {
            $tw->rollback();
            error('Failed to assign activity mangement capability for shared activity creator role');
        }
    }

    $tw->commit();

    return $id;
}

/**
 * Gets the list of supported module names. Adding modules 
 * requires implementation work in several functions within this file. 
 * @return array Array of module names ('forum', etc)
 */
function sharedactv_get_supported_module_names() {
    return array('ouwiki','forum');
}

/**
 * Of the supported module names, return the list that are actually avilable
 * on the system. Cached so it only ever adds one db query.
 * @return Array of module names (zero-length if none)
 */
function sharedactv_get_installed_module_names() {
    static $result=false;
    if($result===false) {
        $in='';
        foreach(sharedactv_get_supported_module_names() as $name) {
            if($in!='') {
                $in.=',';
            }
            $in.="'$name'"; // Assume module names won't contain '        
        }
        $modules=get_records_select('modules',"name IN ($in) AND visible=1",'id,name');
        $modules=$modules ? $modules : array();
        $result=array();
        foreach($modules as $module) {
            $result[]=$module->name;
        }
    }
    return $result;
}

/**
 * Gets the name of the field in the module's table used to store the 'summary'
 * or description.
 *
 * @param string $name Name of module type, e.g. 'forum'
 * @return string Name of field in module's table
 */
function sharedactv_get_module_summary_field($name) {
    switch($name) {
        case 'ouwiki' : return 'summary';
        case 'forum': return 'intro';
        default: error("Unsupported module type $name");
    }
}

function sharedactv_internal_sortbyname($a,$b) {
    return strcasecmp($a->name,$b->name);
}

/**
 * Obtains list of shared activities for the user. This includes both the
 * activities they have created and the ones they've been invited to.
 * The result is an array with two elements keyed SHAREDACTV_ROLE_xx, each
 * of which is an array of $cm objects sorted by $name (activity name) and
 * with $moduletype (e.g. 'forum') as well.
 * @param $userid User ID if specified, otherwise uses urrent user
 * @return array Array as described
 */
function sharedactv_get_shared_activities($userid=null) {
    global $CFG,$USER;
    if(!$userid) {
        $userid=$USER->id;
    }

    // Get roles, course, supported modules
    $roles=sharedactv_get_participant_roles();
    $course=sharedactv_get_magic_course();
    $supportedmodules=sharedactv_get_installed_module_names();

    // Build supported module elements for query
    $supportedmoduleselect='';
    $supportedmodulejoin='';
    $supportedmodulelist='';
    $first=true;
    foreach($supportedmodules as $name) {
        if(!$first) {
            $supportedmodulelist.=',';
            $supportedmoduleselect.=',';
        }
        // Note that this left join does not have a
        $supportedmodulejoin.="LEFT JOIN {$CFG->prefix}$name ON cm.instance={$CFG->prefix}$name.id\n";
        $supportedmoduleselect.="{$CFG->prefix}$name.name AS {$name}_name";
        $supportedmodulelist.="'$name'";
        $first=false;
    }

    // Now let's find all the activities that this user has those roles on.
    if(!($rs=get_recordset_sql($sql="
SELECT
    ra.roleid,m.name AS moduletype,cm.*,$supportedmoduleselect
FROM
    {$CFG->prefix}role_assignments ra
    INNER JOIN {$CFG->prefix}context x ON x.id=ra.contextid
    INNER JOIN {$CFG->prefix}course_modules cm ON cm.id=x.instanceid
    INNER JOIN {$CFG->prefix}modules m ON cm.module=m.id
    $supportedmodulejoin
WHERE
    ra.userid = {$userid}
    AND ra.roleid IN ({$roles[SHAREDACTV_ROLE_CREATOR]},{$roles[SHAREDACTV_ROLE_PARTICIPANT]})
    AND x.contextlevel=".CONTEXT_MODULE."
    AND cm.course = {$course->id}
    AND m.name IN ($supportedmodulelist)
    "))) {
        error('Failed to obtain list of shared activities from database');
    }

    // Organise results by whether you're creator or participant
    $result=array(
        SHAREDACTV_ROLE_CREATOR=>array(),
        SHAREDACTV_ROLE_PARTICIPANT=>array());
    while($rec=rs_fetch_next_record($rs)) {
        // Extract course-module object by removing all the inappropriate fields
        // Leave in the activity type 'moduletype' (nonstandard)
        $cm=clone($rec);
        unset($cm->roleid);
        foreach(array_keys((array)$cm) as $key) {
            if(preg_match('/_name$/',$key)) {
                unset($cm->{$key});
            }
        }

        // Store activity name in there, though, in standard way
        $cm->name=$rec->{$rec->moduletype.'_name'};

        // Save in array, indexed by course-module id
        $result[
            $rec->roleid==$roles[SHAREDACTV_ROLE_CREATOR]
                ? SHAREDACTV_ROLE_CREATOR : SHAREDACTV_ROLE_PARTICIPANT
            ][$cm->id]=$cm;
    }
    rs_close($rs);

    // Sort by activity name and result
    uasort($result[SHAREDACTV_ROLE_CREATOR],'sharedactv_internal_sortbyname');
    uasort($result[SHAREDACTV_ROLE_PARTICIPANT],'sharedactv_internal_sortbyname');
    return $result;
}

/**
 * Obtains information about a single activity based on a course-module ID.
 * @param int $cmid
 * @return object Object containing following fields: $name, $summary, 
 *     $moduletype
 */
function sharedactv_get_activity_info($cmid) {
    global $CFG;

    // Build supported module elements for query
    $supportedmoduleselect='';
    $supportedmodulejoin='';
    foreach(sharedactv_get_installed_module_names() as $name) {
        $summaryfield=sharedactv_get_module_summary_field($name);
        $supportedmoduleselect.=
            ", $name.name AS {$name}_name, $name.$summaryfield AS {$name}_summary";
        $supportedmodulejoin.=
            " LEFT JOIN {$CFG->prefix}$name $name ON m.name='$name' AND cm.instance=$name.id";
    }
    
    $record=get_record_sql("
SELECT
    m.name
    $supportedmoduleselect
FROM
    {$CFG->prefix}course_modules cm
    INNER JOIN {$CFG->prefix}modules m ON cm.module=m.id
    $supportedmodulejoin
WHERE
    cm.id=$cmid");
    if(!$record) {
        error("No such activity $cmid");
    }
    
    $info=new StdClass;
    $info->moduletype=$record->name;
    $namefield=$record->name.'_name';
    $summaryfield=$record->name.'_summary';
    $info->name=$record->{$namefield};
    $info->summary=$record->{$summaryfield};
    $info->id=$cmid;
    return $info;    
}

/**
 * Sets information for an existing activity.
 * @param int $cmid
 * @param string $name Name for activity [no slashes]
 * @param string $summary Text of summary for activity [no slashes]
 */
function sharedactv_set_activity_info($cmid,$name,$summary) {
    global $CFG;
    
    // Get $cm and check this module is on the shared-activity course
    $cm=get_record('course_modules','id',$cmid);
    if(!$cm) {
        error("Invalid course-module: $cmid");
    }
    $course=sharedactv_get_magic_course();
    if($cm->course!=$course->id) {
        error("Not a shared activity: $cmid");
    }
    
    // Handle based on module type
    $mod=new StdClass;
    $mod->id=$cm->instance;
    $mod->name=addslashes($name);
    $moduletype=get_field('modules','name','id',$cm->module);
    switch($moduletype) {
        case 'forum' :
            $mod->intro=addslashes($summary);
            break;
        case 'ouwiki' :
            $mod->summary=addslashes($summary);
            break;
        default :
            error('sharedactv_set_activity_info with unsupported activity type: '.$moduletype);
    }
    
    if(!update_record($moduletype,$mod)) {
        error('Failed to update activity info');
    }
}

/**
 * Creates a shared activity for the current user.
 * @param string $moduletype Type of module e.g. 'forum'
 * @param string $name Name for activity [no slashes]
 * @param string $summary Text of summary for activity [no slashes]
 * @param int $userid User creating activity (0=current)
 * @return int ID of course-module for activity
 */
function sharedactv_create_activity($moduletype,$name,$summary,$userid=0) {
    global $CFG,$USER;
    if(!$userid) {
        $userid=$USER->id;
    }

    // Set up basic module object
    $mod=new StdClass;
    $mod->name=addslashes($name);
    $mod->course=sharedactv_get_magic_course()->id;
    $mod->groupmode=0;
    switch($moduletype) {
        case 'forum' :
            $mod->type='general';
            $mod->intro=addslashes($summary);
            $mod->forcesubscribe=0;
            $mod->cmidnumber=0;
            break;
        case 'ouwiki' :
            $mod->summary=addslashes($summary);
            $mod->subwikis=0;
            break;
        default :
            error('sharedactv_create_activity called with unsupported activity type');
    }

    // Read in module file
    require_once($CFG->dirroot.'/mod/'.$moduletype.'/lib.php');

    // Start transaction
    $tw=new transaction_wrapper();

    // Add object
    $function=$moduletype.'_add_instance';
    $modid=$function($mod);
    if(!$modid) {
        $tw->rollback();
        error('Failed to create shared activity');
    }

    // Add course-modules entry
    $cm=new StdClass;
    $cm->instance=$modid;
    $cm->course=$mod->course;
    $cm->module=get_field('modules','id','name',$moduletype);
    $cmid=insert_record('course_modules',$cm);
    if(!$cmid) {
        $tw->rollback();
        error('Failed to create course-modules entry for shared activity');
    }

    // Add new role assignment
    $roles=sharedactv_get_participant_roles();
    $context=get_context_instance(CONTEXT_MODULE,$cmid);
    if(!role_assign($roles[SHAREDACTV_ROLE_CREATOR],$userid,0,$context->id)) {
        $tw->rollback();
        error('Failed to assign role for shared activity');
    }

    // Finish transaction and return the ID of the module
    $tw->commit();
    return $cmid;
}

/**
 * Deletes an activity and everything associated (hopefully).
 * @param object $cm Course-module object (requires id, course, module, instance)
 */
function sharedactv_delete_activity($cm) {
    global $CFG;

    // Do basic checks
    if($cm->course!=sharedactv_get_magic_course()->id) {
        error('Activity is not on the shared-activity course!');
    }
    $moduletype=get_field('modules','name','id',$cm->module);
    if(!in_array($moduletype,sharedactv_get_supported_module_names())) {
        error('sharedactv_delete_activity called with unsupported activity type');
    }

    // Read in module file
    require_once($CFG->dirroot.'/mod/'.$moduletype.'/lib.php');

    // Start transaction
    $tw=new transaction_wrapper();

    // Delete activity
    $function=$moduletype.'_delete_instance';
    $ok=$function($cm->instance);
    if(!$ok) {
        $tw->rollback();
        error('Failed to delete shared activity');
    }

    // Delete course_modules line
    $ok=delete_course_module($cm->id);
    if(!$ok) {
        $tw->rollback();
        error('Failed to delete shared activity course-modules entry');
    }
    
    // Delete context and roles
    if(!delete_context(CONTEXT_MODULE,$cm->id)) {
        $tw->rollback();
        error('Failed to delete context and roles');
    }

    // Finish transaction
    $tw->commit();
}

/**
 * Adds a participant to an existing shared activity.
 * @param int $cmid Course-module ID of activity
 * @param int $userid ID of user to add
 * @return True if successful
 */
function sharedactv_add_participant($cmid,$userid) {
    // Add new role assignment
    $roles=sharedactv_get_participant_roles();
    $context=get_context_instance(CONTEXT_MODULE,$cmid);
    return role_assign($roles[SHAREDACTV_ROLE_PARTICIPANT],$userid,0,
        $context->id,0,0,0,SHAREDACTV_ENROLTYPE);
}

/**
 * Removes a participant from an existing shared activity.
 * @param int $cmid Course-module ID of activity
 * @param int $userid ID of user to remove
 * @return True if successful
 */
function sharedactv_remove_participant($cmid,$userid) {
    // Delete existing role assignment
    $roles=sharedactv_get_participant_roles();
    $context=get_context_instance(CONTEXT_MODULE,$cmid);
    return role_unassign($roles[SHAREDACTV_ROLE_PARTICIPANT],$userid,0,
        $context->id,SHAREDACTV_ENROLTYPE);
}

/**
 * Obtains the list of all participants on a shared activity.
 * @param int $cmid Course-module ID of activity
 * @param string $type If set to SHAREDACTV_ROLE_xx constant, returns only
 *   a single array for that type
 * @return Array of SHAREDACTV_ROLE_xx => array of user objects
 *   (with id, username, firstname, lastname, keyed on id)
 */
function sharedactv_get_participants($cmid,$type=null) {
    global $CFG;

    // Get list of users, ordered by name
    $context=get_context_instance(CONTEXT_MODULE,$cmid);
    if(!($rs=get_recordset_sql("
SELECT
    ra.roleid,u.id,u.firstname,u.lastname,u.username
FROM
    {$CFG->prefix}role_assignments ra
    INNER JOIN {$CFG->prefix}user u ON u.id=ra.userid
WHERE
    ra.contextid={$context->id}
ORDER BY
    u.lastname,u.firstname
    "))) {
        error('Failed to obtain list of participants from database');
    }

    // Organise results by whether you're creator or participant
    $roles=sharedactv_get_participant_roles();
    $result=array(
        SHAREDACTV_ROLE_CREATOR=>array(),
        SHAREDACTV_ROLE_PARTICIPANT=>array());
    while($rec=rs_fetch_next_record($rs)) {
        $user=clone($rec);
        unset($user->roleid);
        if($rec->roleid == $roles[SHAREDACTV_ROLE_CREATOR]) {
            $result[SHAREDACTV_ROLE_CREATOR][$user->id]=$user;
        } else if($rec->roleid == $roles[SHAREDACTV_ROLE_PARTICIPANT]) {
            $result[SHAREDACTV_ROLE_PARTICIPANT][$user->id]=$user;
        }
    }
    
    // Split out only wanted roles if specified
    if($type) {
        return $result[$type];
    } else {
        return $result;
    }
}

/**
 * Displays the 'view other' bar that goes at top of pages to indicate 
 * you're viewing somebody else's activities, if you are!
 *
 * @param object $user The person whose activities you're viewing
 */
function sharedactv_ui_viewotherbar($user) {
    global $USER;
    if($user->id==$USER->id) {
        return;
    }
    $strviewing=get_string('viewing','format_sharedactv',
        htmlspecialchars(fullname($user)).' ('.
        htmlspecialchars($user->username).')');
    $strbacktome=get_string('backtome','format_sharedactv');
        
    print <<<END
<form action="viewother.php" method="post" class="sa_viewingother"><div>
<input type="hidden" name="username" value="" />
$strviewing
<input type="submit" value="$strbacktome" />
</div></form>
END;
}

/**
 * Logs an action within the shared activities system.
 *
 * @param object $course Course object
 * @param string $action Action type
 * @param int $cmid
 * @param string $info Optional additional info
 */
function sharedactv_log($course,$action,$cmid,$info='') {
    global $CFG,$USER;
    add_to_log($course->id,'sharedactv',$action,
        $CFG->wwwroot.'/course/format/sharedactv/go.php?to='.$cmid,
        'Activity: '.$cmid.' '.$info);
}

?>